package main

import (
	"bufio"
	"bytes"
	"encoding/base64"
	"errors"
	"fmt"
	"io"
	"io/fs"
	"log"
	"math/rand"
	"net"
	"os"
	"os/signal"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"syscall"
	"time"

	"github.com/tobischo/gokeepasslib/v3"
	"ser1.net/claptrap/v4"
)

type server struct {
	db         *gokeepasslib.Database
	dbPath     string
	timestamp  time.Time
	closer     chan bool
	sock       net.Listener
	socketPath string
	pin        string
	otpin      int
	attempts   int
	locked     bool
}

func new(root *claptrap.CommandConfig, keydata []byte) server {
	dbPath := root.Command.String("database")
	// If the server was not started with a database, start in stand-by and wait for a client call to open one
	var db *gokeepasslib.Database
	var err error
	if dbPath != "" {
		db = gokeepasslib.NewDatabase()
		if root.String("password") != "" && len(keydata) == 0 {
			db.Credentials = gokeepasslib.NewPasswordCredentials(root.String("password"))
		} else if root.String("password") != "" {
			if db.Credentials, err = gokeepasslib.NewPasswordAndKeyDataCredentials(root.String("password"), keydata); err != nil {
				log.Fatalf("error decoding the key data: %s", err)
			}
		}
		// else, the client will have to "open" the database with a password later
	}

	var sock net.Listener
	sockPath := socketPath()
	if _, err = os.Stat(sockPath); !os.IsNotExist(err) {
		if root.Bool("force") {
			os.Remove(sockPath)
		} else if s, err := net.Dial("unix", socketPath()); err == nil {
			s.Close()
			log.Fatal("pretty sure a server is already be running. If you are certain it isn't, run again with -f. If you do this, and you're wrong, something is going to break.")
		} else {
			os.Remove(sockPath)
		}
	}
	syscall.Umask(0077)
	if sock, err = net.Listen("unix", sockPath); err != nil {
		log.Fatalf("unexpected error setting up server socket: %s", err)
	}

	server := server{
		db:     db,
		dbPath: dbPath,
		sock:   sock,
		closer: make(chan bool),
		otpin:  -1,
	}
	if server.db != nil && server.db.Credentials != nil {
		if err := server.reload(); err != nil {
			log.Printf("[ERROR] reload() error on server creation")
			server.locked = true
		}
	}

	return server
}

// start launches a rook server. This function does not return until the server
// has been closed by a client, or an unrecoverable error has been encountered.
func (srvr server) start() {
	ctlc := make(chan os.Signal, 1)
	signal.Notify(ctlc, os.Interrupt)

	defer log.Print("[INFO] closed and cleaned")
	defer os.Remove(srvr.socketPath)
	defer srvr.sock.Close()

	clients := make(chan net.Conn, 5)
	go func() {
		for {
			conn, err := srvr.sock.Accept()
			if err != nil {
				if !errors.Is(err, net.ErrClosed) {
					log.Print("[ERROR] accept error: ", err)
				}
				return
			}
			clients <- conn
		}
	}()

	fmt.Fprintf(os.Stderr, "Rook server started with %q\n", srvr.dbPath)
	if srvr.pin != "" {
		srvr.pin = fmt.Sprintf("%04d", rand.Intn(99_999))
		fmt.Println(srvr.pin)
		// Close this so we don't hang up any pipes
		os.Stdout.Close()
	}
	for {
		select {
		case conn := <-clients:
			go srvr.process(conn)
		case <-srvr.closer:
			return
		case <-ctlc:
			return
		}
	}
}

// The wire protocol is simple:
// COMMAND\rkey1\rvalue1\r...\0
// The \r on the last value is optional
// An entire single message may not exceed 2048 bytes. If this is exceeded, a FAIL message is immediately sent and message parsing resets.
// If a single message exceeds 1MB, the socket is closed and reset.
// Arguments typically take the form `key=value`
//
// Responses are in the same format: multiple values are returned separated by \r, records are separated by double-\r, and messages are terminated by \0.
//
// TODO (D) But *sigh* the wire format should probably be json or some shit
// [cbor](https://github.com/brianolson/cbor_go) looks awesome; it'd add an external dependency and 1155 more lines to audit.
func (srvr *server) process(conn net.Conn) {
	defer conn.Close()
	rdr := bufio.NewReader(io.LimitReader(conn, 2048))
	line, err := rdr.ReadBytes('\000')
	if err != nil {
		if err == io.EOF {
			log.Printf("[ERROR] message too long error; message was")
			log.Printf("%q", string(line))
			conn.Write([]byte(fmt.Sprintf("FAIL\fMALFORMED MESSAGE\000")))
		}
		log.Printf("[ERROR] error encountered parsing string")
	}
	line = bytes.TrimRight(bytes.TrimSpace(line), "\000")
	parts := strings.Split(string(line), "\r")
	if len(parts) < 1 {
		log.Printf("[ERROR] empty message")
		conn.Write([]byte(fmt.Sprintf("FAIL\fEMPTY MESSAGE\000")))
	}
	command := parts[0]
	if command == STATUS {
		srvr.status(conn)
		conn.Write([]byte{'\000'})
		return
	}
	parts = parts[1:]
	if srvr.pin != "" && command != LOCK {
		if command != OPEN {
			srvr.attempts += 1
			if len(parts) < 1 || parts[0] != srvr.pin {
				log.Printf("[ERROR] auth fail %d #%d", len(parts), srvr.attempts)
				if srvr.checkFailAttempts() {
					write(conn, "FAIL\fAUTH TRIES EXCEEDED\fDB HARD LOCKED\000")
				} else {
					write(conn, "FAIL\fAUTH\000")
				}
				return
			}
			srvr.attempts = 0
			parts = parts[1:]
		}
	}
	if !COMMANDS[command] {
		msg := fmt.Sprintf("[ERROR] unknown command: %q (%d)", command, len(command))
		conn.Write([]byte(fmt.Sprintf("FAIL\fUNKNOWN COMMAND\000")))
		log.Println(msg)
		return
	}
	args := make(map[string]string)
	if len(parts) > 1 {
		for i := 0; i < len(parts); i += 2 {
			var val string
			if i+1 < len(parts) {
				val = parts[i+1]
			}
			args[parts[i]] = val
		}
	}
	log.Printf("[DEBUG] args are %#v", args)
	defer conn.Write([]byte{'\000'})
	if command == LOCK && args[OTPIN] == TRUE {
		// This could be an attempt to bypass the pin-lock, but we'll soft-fail it
		// anyway since it doesn't hurt anything.
		if srvr.otpin != -1 {
			log.Printf("[WARN] attempted pin re-lock of locked DB fail")
			conn.Write([]byte("FAIL\fPIN-LOCKED DB"))
			return
		}
		srvr.otpin = int(rand.Int31n(10000))
		srvr.attempts = 0
		log.Print("[INFO] one-time pin locking DB")
		write(conn, "%04d", srvr.otpin)
		return
	}
	if srvr.otpin != -1 && !(command == OPEN && args[PASSWORD] != "") {
		// We do not count attempts to access pin-locked databases, only failed
		// attempts to unlock
		if command != OPEN {
			log.Printf("[WARN] attempted access of pin-locked DB; soft fail")
			conn.Write([]byte("FAIL\fPIN-LOCKED DB"))
			return
		}
		srvr.attempts += 1
		if pin, ok := strconv.Atoi(args[OTPIN]); ok != nil || pin != srvr.otpin {
			log.Printf("[DEBUG] parsed (err: %v) pin provided (%d) == otpin? %t", ok, pin, pin == srvr.otpin)
			if srvr.checkFailAttempts() {
				write(conn, "FAIL\fPIN UNLOCK TRIES EXCEEDED")
				return
			}
			log.Printf("[WARN] incorrect pin")
			write(conn, "FAIL\fINCORRECT ONE-TIME PIN")
			return
		} else {
			srvr.otpin = -1
			srvr.attempts = 0
			log.Print("[INFO] one-time pin unlock succeeded")
			write(conn, "SUCCESS")
			return
		}
	}
	if command != LOCK && command != OPEN && command != LISTCMDS {
		if srvr.db == nil {
			log.Println("[WARN] no open database")
			conn.Write([]byte(fmt.Sprintf("FAIL\fNO OPEN DATABASE")))
			return
		}
		if command != RELOAD && srvr.needsRefresh() {
			if err := srvr.reload(); err != nil {
				log.Printf("[WARN] reload() failed on %s", command)
				conn.Write([]byte(fmt.Sprintf("FAIL\fNO OPEN DATABASE")))
				return
			}
		}
	}
	switch command {
	default:
		err = errors.New("[ERROR] unknown command (Twilight Zone version)")
	case LISTCMDS:
		err = commands(conn)
	case OPEN:
		err = srvr.open(conn, args)
		obfuscate(args)
	case LOCK:
		if err = srvr.lock(); err == nil {
			writeLn(conn, "LOCKED")
		}
	case LIST:
		err = srvr.list(conn, args)
	case EXIT:
		if err = srvr.kill(); err == nil {
			writeLn(conn, "SUCCESS")
		}
	case INFO:
		err = srvr.info(conn)
	case SEARCH:
		err = srvr.search(conn, args)
	case MATCH:
		err = srvr.match(conn, args)
	case RELOAD:
		if err = srvr.reload(); err == nil {
			writeLn(conn, "SUCCESS")
		} else {
			log.Printf("[ERROR] reload() failed on RELOAD command")
		}
	case SHOW:
		err = srvr.show(conn, args)
	}
	if err != nil {
		conn.Write([]byte(fmt.Sprintf("FAIL\f%s\000", err)))
		log.Printf("[ERROR] on %s command: %s\n", command, err)
	} else {
		log.Printf("[INFO] success: %s %v", command, args)
	}
}

func commands(response io.Writer) error {
	for k, v := range COMMANDS {
		if v {
			response.Write([]byte(fmt.Sprintf("%s\r", k)))
		}
	}
	return nil
}

func (srvr *server) open(response io.Writer, args map[string]string) error {
	// the server could have been started with a dbPath, in which case open doesn't
	// need to supply one. However, if the client _does_ supply one, use it instead
	// of what the server was started with.
	dbPath := args[DBPATH]
	if dbPath == "" {
		dbPath = srvr.dbPath
	}
	if _, err := os.Stat(dbPath); err != nil {
		return err
	}
	password := args[PASSWORD]
	//println("trying to open ", dbPath, " with ", password)
	srvr.db = gokeepasslib.NewDatabase()
	if enckey, ok := args[KEYDATA]; !ok {
		srvr.db.Credentials = gokeepasslib.NewPasswordCredentials(password)
	} else {
		keydata, err := base64.StdEncoding.DecodeString(enckey)
		if err != nil {
			return err
		}
		if srvr.db.Credentials, err = gokeepasslib.NewPasswordAndKeyDataCredentials(password, keydata); err != nil {
			log.Fatalf("error decoding the key data: %s", err)
		}
	}
	srvr.dbPath = dbPath
	if err := srvr.reload(); err != nil {
		return err
	}
	if srvr.pin == "XXXX" {
		srvr.pin = fmt.Sprintf("%06d", rand.Intn(99_999))
	}
	if srvr.pin != "" {
		write(response, srvr.pin)
	} else {
		write(response, "SUCCESS")
	}
	srvr.otpin = -1
	srvr.locked = false
	return nil
}

func (srvr *server) lock() error {
	// This is the best we can do, since gokeepasslib offers no API for closing or locking a database
	srvr.db.LockProtectedEntries()
	srvr.db.Credentials = nil
	srvr.db.Content = nil
	srvr.db = nil
	// Reset the pin, but preserve that the user ran in --secure mode
	if srvr.pin != "" {
		srvr.pin = "XXXX"
	}
	srvr.locked = true
	return nil
}

func (srvr server) list(response io.Writer, args map[string]string) error {
	_, ok := args["tree"]
	if ok {
		// tree is a map of child -> parent
		tree := make(map[string][]string)
		tree["/"] = make([]string, 0)
		srvr.forEach(
			func(idx int, parts []string, entry gokeepasslib.Entry) bool {
				parts = append([]string{"/"}, parts...)
				for len(parts) > 0 {
					zName := parts[0]
					if _, ok := tree[zName]; !ok {
						tree[zName] = make([]string, 0)
					}
					parts = parts[1:]
					if len(parts) == 0 {
						tree[zName] = append(tree[zName], entry.GetTitle())
					} else {
						var found bool
						for _, itm := range tree[zName] {
							if itm == parts[0] {
								found = true
								break
							}
						}
						if !found {
							tree[zName] = append(tree[zName], parts[0])
						}
					}
				}
				return true
			},
		)
		writeTree(response, tree)
	} else {
		srvr.forEach(
			func(idx int, pth []string, entry gokeepasslib.Entry) bool {
				response.Write([]byte(fmt.Sprintf("%d\t%s\t%s\r", idx, strings.Join(pth, "/"), entry.GetTitle())))
				return true
			},
		)
	}
	return nil
}

func (srvr server) kill() error {
	log.Print("[INFO] closed by client")
	srvr.closer <- true
	return nil
}

func (srvr server) status(response io.Writer) error {
	if srvr.otpin != -1 {
		response.Write([]byte("pin locked\r"))
		return nil
	}
	if srvr.db == nil || srvr.locked {
		response.Write([]byte("locked\r"))
		return nil
	}
	response.Write([]byte("unlocked\r"))
	return nil
}

func (srvr server) info(response io.Writer) error {
	response.Write([]byte(fmt.Sprintf("Database: %s\r", srvr.dbPath)))
	version := fmt.Sprintf("Database version: %d.%d", srvr.db.Header.Signature.MajorVersion, srvr.db.Header.Signature.MinorVersion)
	response.Write([]byte(version))
	return nil
}

// search finds entries who's title or URL matches a regular expression supplied
// in the args map as "searchTerm". All entries that match are printed.
func (srvr server) search(out io.Writer, args map[string]string) error {
	titleTerm := args[TITLE]
	pathTerm := args[PATH]
	titleRe, terr := regexp.Compile("(?i)" + regexp.QuoteMeta(titleTerm))
	pathRe, perr := regexp.Compile("(?i)" + regexp.QuoteMeta(pathTerm))
	srvr.forEach(
		func(idx int, pth []string, entry gokeepasslib.Entry) bool {
			// Filter on the path. If we have a path, but it didn't compile as an RE,
			// then do a simple string match; otherwise, test the RE. If the current path
			// doesn't match, then we skip this entry.
			if pathTerm != "" {
				path := strings.Join(pth, "/")
				if perr == nil {
					if !pathRe.Match([]byte(path)) {
						return true
					}
				} else if !strings.Contains(path, pathTerm) {
					// Fallback in case there was some error with the regexp compilation
					return true
				}
			}
			var url string
			if entry.Get("URL") != nil {
				url = entry.Get("URL").Value.Content
			}
			if terr == nil {
				if titleRe.Match([]byte(entry.GetTitle())) || titleRe.Match([]byte(url)) || titleRe.Match([]byte(entry.Tags)) {
					printEntry(out, idx, pth, entry)
					out.Write([]byte("\r"))
				}
				return true
			}
			// Fallback in case there was some error with the regexp compilation
			if strings.Contains(entry.GetTitle(), titleTerm) || strings.Contains(url, titleTerm) || strings.Contains(entry.Tags, titleTerm) {
				printEntry(out, idx, pth, entry)
				out.Write([]byte("\r"))
			}
			return true
		},
	)
	return nil
}

// match is the reverse of search, in that -- given a string -- if an entry's
// title, url, or autocomplete pattern matches/is a substring of the string, it
// is a match.
// match returns all matches, in quality order (the highest quality match is first)
func (srvr server) match(out io.Writer, args map[string]string) error {
	type match struct {
		idx      int
		quality  int
		entry    gokeepasslib.Entry
		path     string
		sequence string
	}
	matches := make([]match, 0)
	activeWindow := args[SEARCHTERM1]
	srvr.forEach(
		func(i int, p []string, entry gokeepasslib.Entry) bool {
			q, s := checkAssociations(entry, activeWindow)
			if q > 0 {
				matches = append(matches, match{
					idx:      i,
					quality:  q,
					entry:    entry,
					path:     strings.Join(p, "/"),
					sequence: s,
				})
			}
			return true
		},
	)
	sort.Slice(matches, func(i, j int) bool {
		return matches[i].quality > matches[j].quality
	})
	log.Printf("[DEBUG] match(%q) found %d matches", args[SEARCHTERM1], len(matches))
	for _, match := range matches {
		writeLn(out, "%d\t%s\t%s", match.idx, match.path, match.entry.GetTitle())
		writeLn(out, "%s", match.sequence)
		printEntryFull(out, match.entry, false)
		endRecord(out)
	}
	return nil
}

// checkAssociations checks a window title against an entry's autotype configuration,
// and if it matches, returns a match quality and an autotype sequence
func checkAssociations(entry gokeepasslib.Entry, activeWindow string) (int, string) {
	type tuple struct {
		qua int
		seq string
	}
	defaultSequence := "{USERNAME}{TAB}{PASSWORD}{ENTER}"
	if entry.AutoType.DefaultSequence != "" {
		defaultSequence = entry.AutoType.DefaultSequence
	}
	matches := make([]tuple, 0)
	if entry.AutoType.Associations != nil {
		for _, at := range entry.AutoType.Associations {
			// Check if there's a window association, and if so, make sure it's a regexp
			if at.Window == "" {
				continue
			}
			ms := at.Window
			if !strings.HasPrefix(at.Window, "//") {
				// Replace all star globs with .*
				ms = strings.ReplaceAll(ms, "*", ".*")
			} else {
				// For regexp, remove the wrapping
				if len(ms) > 2 {
					ms = ms[2:]
				}
				if len(ms) > 2 {
					ms = ms[:len(ms)-2]
				}
			}
			reg, err := regexp.Compile(ms)
			if err != nil {
				continue
			}
			if reg.MatchString(activeWindow) {
				// TODO Review this hacky way of evaluting the quality of a match string
				quality := len(ms)
				quality -= strings.Count(ms, "*")
				seq := defaultSequence
				if at.KeystrokeSequence != "" {
					seq = at.KeystrokeSequence
				}
				matches = append(matches, tuple{quality, seq})
			}
		}
	}
	reg, err := regexp.Compile(".*" + entry.GetTitle() + ".*")
	if err == nil {
		if reg.MatchString(activeWindow) {
			// TODO Review this hacky way of evaluting the quality of a match string
			quality := len(entry.GetTitle()) - 2
			matches = append(matches, tuple{quality, defaultSequence})
		}
	}
	switch len(matches) {
	case 0:
		return -1, ""
	case 1:
		return matches[0].qua, matches[0].seq
	default:
		sort.Slice(matches, func(i, j int) bool {
			return matches[i].qua < matches[j].qua
		})
		return matches[0].qua, matches[0].seq
	}
}

func printEntry(out io.Writer, idx int, pth []string, entry gokeepasslib.Entry) {
	out.Write([]byte(fmt.Sprintf("Index: %d\rPath: %s\rTitle: %s\r\f", idx, strings.Join(pth, "/"), entry.GetTitle())))
}
func printEntryFull(out io.Writer, entry gokeepasslib.Entry, protect bool) {
	for _, value := range entry.Values {
		val := "******"
		if !value.Value.Protected.Bool || !protect {
			val = value.Value.Content
		}
		writeLn(out, "%s: %s", value.Key, val)
	}
	if !protect {
		otp, err := CreateOTP(entry, time.Now().Unix())
		if err != nil {
			return
		}
		writeLn(out, "totp: %s", otp)
	}
}
func write(out io.Writer, format string, args ...any) {
	out.Write([]byte(fmt.Sprintf(format, args...)))
}
func writeLn(out io.Writer, format string, args ...any) {
	out.Write([]byte(fmt.Sprintf(format, args...)))
	out.Write([]byte("\r"))
}
func endRecord(out io.Writer) {
	out.Write([]byte("\f"))
}

// show lists all of the populated fields in a given record, or individual field entries
func (srvr server) show(out io.Writer, args map[string]string) error {
	write := func(idx int, path string, entry gokeepasslib.Entry) {
		// allKeys is a map of all keys converted to upper case, mapping to their actual key strings
		allKeys := make(map[string]string)
		for _, value := range entry.Values {
			allKeys[strings.ToUpper(value.Key)] = value.Key
		}
		fieldU := strings.ToUpper(args["field"])
		if fieldU == "PASSWORD" {
			write(out, "%s", entry.GetPassword())
		} else if fieldU == "USERNAME" {
			write(out, "%s", entry.Get("UserName").Value.Content)
		} else if fieldU == "OTP" {
			otp, err := CreateOTP(entry, time.Now().Unix())
			if err != nil {
				write(out, "ERROR GENERATING OTP")
			} else {
				write(out, otp)
			}
		} else if fieldU != "" {
			f := allKeys[fieldU]
			if entry.Get(f) != nil {
				write(out, "%s", entry.Get(f).Value.Content)
			}
		} else {
			writeLn(out, "%d\t%s\t%s", idx, path, entry.GetTitle())
			printEntryFull(out, entry, true)
		}
		endRecord(out)
	}
	path := strings.ToLower(args[PATH])
	term := strings.ToLower(args[TITLE])
	if term != "" {
		srvr.forEach(
			func(idx int, pth []string, entry gokeepasslib.Entry) bool {
				pathJoined := strings.Join(pth, "/")
				loweredPath := strings.ToLower(pathJoined)
				if !strings.Contains(loweredPath, path) {
					return true
				}
				ttl := strings.ToLower(entry.GetTitle())
				if strings.Contains(ttl, term) {
					write(idx, pathJoined, entry)
				}
				return true
			},
		)
	} else if args["index"] != "" {
		index, _ := strconv.Atoi(args["index"])
		srvr.forEach(
			func(idx int, pth []string, entry gokeepasslib.Entry) bool {
				if idx == index {
					path := strings.Join(pth, "/")
					write(idx, path, entry)
				}
				return true
			},
		)
	}
	return nil
}

func (srvr server) reload() (err error) {
	var st fs.FileInfo
	defer func() {
		srvr.locked = err == nil
	}()
	st, err = os.Stat(srvr.dbPath)
	if err != nil {
		log.Printf("[ERROR] error while accessing database file %q: %s", srvr.dbPath, err)
		return
	}
	if fin, err := os.Open(srvr.dbPath); err == nil {
		defer fin.Close()
		if err = gokeepasslib.NewDecoder(fin).Decode(srvr.db); err == nil {
			if err = srvr.db.UnlockProtectedEntries(); err != nil {
				log.Printf("[ERROR] unable to unlock database: %s", err)
				return err
			}
		} else {
			log.Printf("[ERROR] unable to decode database %q: %s", srvr.dbPath, err)
			return err
		}
	} else {
		log.Printf("[ERROR] error while accessing database %q: %s", srvr.dbPath, err)
		return err
	}
	srvr.timestamp = st.ModTime()
	return nil
}

// forEach recursively decends the database tree, depth-first, and calls the
// supplied function for each entry. The (group) path to the entry is provided as well.
// If the function returns true, the recursion
// continues; if false, the process is halted and the function returns immediately.
// This will get more exciting with coroutines.
// https://github.com/golang/go/issues/61897
func (srvr server) forEach(fn func(int, []string, gokeepasslib.Entry) bool) {
	// Protect against closures in the middle of processing
	db := srvr.db
	if db == nil {
		log.Printf("[WARN] no open database")
		return
	}
	var iterGroup func([]string, gokeepasslib.Group)
	var idx int
	iterGroup = func(pth []string, group gokeepasslib.Group) {
		if group.Groups != nil {
			for _, sub := range group.Groups {
				subpth := append(pth, sub.Name)
				iterGroup(subpth, sub)
			}
		}
		if group.Entries != nil {
			for _, entry := range group.Entries {
				idx++
				if !fn(idx, pth, entry) {
					return
				}
			}
		}
	}
	if db.Content != nil && db.Content.Root != nil {
		for _, group := range db.Content.Root.Groups {
			iterGroup([]string{}, group)
		}
	}
}

func (srvr server) needsRefresh() bool {
	st, err := os.Stat(srvr.dbPath)
	return err == nil && st.ModTime().After(srvr.timestamp)
}

func (srvr *server) checkFailAttempts() bool {
	if srvr.attempts > MAX_PIN_UNLOCK_ATTEMPTS {
		log.Printf("[ERROR] too many one-time pin unlock attempts... hard-locking DB")
		srvr.attempts = 0
		srvr.otpin = -1
		srvr.pin = "XXXX"
		srvr.lock()
		return true
	}
	return false
}

func obfuscate(args map[string]string) {
	for _, key := range []string{PASSWORD, KEYDATA, OTPIN} {
		if _, ok := args[key]; ok {
			args[key] = "******"
		}
	}
}
